<!-- SPDX-License-Identifier: BSD-3-Clause -->
<!-- Copyright Contributors to the OpenEXR Project. -->

{% extends '!layout.html' %}

{% block footer %}

<p><center><img height="30" width="160" src="https://artwork.aswf.io/other/aswf/logo/black/aswf-logo-black.svg" /></center></p>

<p><small><center>Copyright © OpenEXR a Series of LF Projects, LLC. </center></small></p>
<p><small><center>For web site terms of use, trademark policy and other project policies
please see <a href="https://lfprojects.org">https://lfprojects.org</a>.</center></small></p>

{% endblock %}
