//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Contributors to the OpenEXR Project.
//

#ifndef TESTMULTIPARTSHAREDATTRIBUTES_H_
#define TESTMULTIPARTSHAREDATTRIBUTES_H_

#include <string>

void testMultiPartSharedAttributes (const std::string& tempDir);

#endif /* TESTMULTIPARTSHAREDATTRIBUTES_H_ */
