// [Rgba definition begin]
struct Rgba
{
    half r; // red
    half g; // green
    half b; // blue
    half a; // alpha (opacity)
};
// [Rgba definition end]

// [GZ definition begin]
struct GZ
{
    half g;
    float z;
};
// [GZ definition end]