// SPDX-License-Identifier: BSD-3-Clause
// Copyright Contributors to the OpenEXR Project.

#ifndef OPENEXR_CORE_TEST_DEEP_H
#define OPENEXR_CORE_TEST_DEEP_H

#include <string>

void testOpenDeep (const std::string& tempdir);

void testReadDeep (const std::string& tempdir);
void testWriteDeep (const std::string& tempdir);

#endif // OPENEXR_CORE_TEST_READ_H
