//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Contributors to the OpenEXR Project.
//

#ifndef TESTMULTIPARTTHREADING_H_
#define TESTMULTIPARTTHREADING_H_

#include <string>

void testMultiPartThreading (const std::string& tempDir);

#endif /* TESTMULTIPARTTHREADING_H_ */
