//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Contributors to the OpenEXR Project.
//

#ifndef TESTMULTISCANLINEPARTTHREADING_H_
#define TESTMULTISCANLINEPARTTHREADING_H_

#include <string>

void testMultiScanlinePartThreading (const std::string& tempDir);

#endif /* TESTMULTISCANLINEPARTTHREADING_H_ */
