//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Contributors to the OpenEXR Project.
//

#include <string>

void testExistingStreams (const std::string& tempDir);
void testExistingStreamsUTF8 (const std::string& tempDir);
