//
// SPDX-License-Identifier: BSD-3-Clause
// Copyright (c) Weta Digital, Ltd and Contributors to the OpenEXR Project.
//

#ifndef TESTDEEPSCANLINEMULTIPLEREAD_H_
#define TESTDEEPSCANLINEMULTIPLEREAD_H_

#include <string>

void testDeepScanLineMultipleRead (const std::string& tempDir);

#endif /* DEEPSCANLINEMULTIPLEREAD_H_ */
